using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.EventSystems;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using MeshVR;
using SimpleJSON;
using MacGruber;

/// <summary>
/// This plugin is about editing GameObjects (enabling/disabling/deleting) in CUAs.
/// Based on a mix of ideas of CUAshaderMod by Sally Whitemane and DisableCUAObject2 by unique.
/// 
/// Version history:
///   1.0: Initial version
///   1.1: - Added audio sources as type
///        - Moved destruction UI elements a bit + set to red
///        - Fixed the bug that a type filter change leads to an empty UI
///        - Fixed that ugly type enumeration/index code
///        - Need to restructure source directory as MacGruber_Utils need to come on board for filter by name (breaking change)
///        - Added filter by name
///        - Added output to clipboard when printing the structure
///        - Added functions to save/restore configuration to/from clipboard
///   1.2: - moved Rescan and Restore buttons back into a proper order 
///   1.3: - Adding prefixes to be able to deal with identical atom names
///   1.4: - Adding hover-over-toggle-to-hilight-object functionality
///   1.5: - Bugfix for parented CUAs
/// </summary>

namespace HAL9001
{
	public class CUA_Editor : MVRScript
	{
		public static readonly string PLUGIN_VERSION = "1.5";

		// enumeration for filtering by type
		public static readonly string TYPE_GAMEOBJECTS      = "Game Objects";
		public static readonly string TYPE_GAMEOBJECT_ID    = "G";
		public static readonly string TYPE_RENDERERS        = "Renderers";
		public static readonly string TYPE_RENDERER_ID      = "R";
		public static readonly string TYPE_COLLIDERS        = "Colliders";
		public static readonly string TYPE_COLLIDER_ID      = "C";
		public static readonly string TYPE_LIGHTS           = "Lights";
		public static readonly string TYPE_CAMERAS          = "Cameras";
		public static readonly string TYPE_AUDIOSOURCES     = "Audio Sources";
		public static readonly string TYPE_AUDIOSOURCE_ID   = "A";
		public static readonly string TYPE_LIGHTPROBES      = "Light Probes";
		public static readonly string TYPE_REFLECTIONPROBES = "Reflection Probes";
		public static readonly List<string>  TYPE_OPTIONS   = new List<string>  {TYPE_GAMEOBJECTS, TYPE_RENDERERS, TYPE_COLLIDERS, TYPE_LIGHTS, TYPE_CAMERAS, TYPE_AUDIOSOURCES, TYPE_LIGHTPROBES, TYPE_REFLECTIONPROBES};

		// colours
		public static readonly Color32 COLOUR_ENABLED     = new Color32(192, 255, 192, 255); // green
		public static readonly Color32 COLOUR_DISABLED    = new Color32(255, 255, 192, 255); // yellow
		public static readonly Color32 COLOUR_DISABLED_GO = new Color32(240, 255, 192, 255); // pale green/yellow
		public static readonly Color32 COLOUR_DESTROYED   = new Color32(255, 128, 128, 255); // red


		protected class CUA_GameObject 
		{
			public CUA_Editor         script;
			public string             prefix, name, preV13name;
			public bool               destroyed;
			public GameObject         gameObject;
			public JSONStorableBool   gameObjectToggle;
			public Bounds             bounds;

			public Renderer[]         renderers;
			public JSONStorableBool[] rendererToggles;
			
			public Collider[]         colliders;
			public JSONStorableBool[] colliderToggles;

			public AudioSource[]      audioSources;
			public JSONStorableBool[] audioSourceToggles;
			
			public Light              light;

			public Camera             camera;
			public LightProbeGroup    lightProbeGroup;
			public ReflectionProbe    reflectionProbe;

			protected List<JSONStorableBool> m_toggles;
			protected List<string>           m_preV13toggleNames;
			protected List<UIDynamicToggle>  m_uiElements;
			protected List<Color>            m_colors;


			public CUA_GameObject(CUA_Editor script, Transform transform, string prefix)
			{
				this.script     = script;
				this.gameObject = transform.gameObject;
				this.destroyed  = false;
				this.prefix     = prefix;
				
				name = gameObject.name;

				// some name cleanup (pre v1.3 version, to allow for backwards compatibility)
				preV13name = name.Replace("(Clone)", "");
				preV13name = name.Replace("____", "_").Replace("___", "_").Replace("__", "_");

				// some name cleanup
				name = Regex.Replace(name, "^[\\s_]+", "");  // remove underscores at beginning
				name = Regex.Replace(name, "[\\s_]+$", "");  // remove underscores at end
				name = Regex.Replace(name, "_{2,}",  "_");   // collapse multiple underscores
				name = Regex.Replace(name, "\\s{2,}", " ");  // collapse multiple whitespaces
				name = Regex.Replace(name, "[_ ]*\\(Clone\\)[_ ]*", ""); // remove "(Clone)" postfixes
				name = Regex.Replace(name, "\\s\\((\\d+)\\)", "_${1}");  // replace Unity automatic indices (e.g., " (123)" by "_123")
				name = Regex.Replace(name, "['\"]*", "");                // remove quotation marks

				m_uiElements        = new List<UIDynamicToggle>();
				m_toggles           = new List<JSONStorableBool>();
				m_preV13toggleNames = new List<string>();

				m_colors = new List<Color>();

				DetermineComponents();
				DetermineBounds();
			}


			public void DetermineComponents()
			{
				if (gameObject != null)
				{
					string prefixedName = string.IsNullOrEmpty(prefix) ? name : (prefix + "_" + name);

					int counter;
					gameObjectToggle = new JSONStorableBool(TYPE_GAMEOBJECT_ID + "_" + prefixedName, gameObject.activeSelf, (bool _val) => UpdateGameObject());
					m_toggles.Add(gameObjectToggle);
					m_preV13toggleNames.Add(TYPE_GAMEOBJECT_ID + "_" + preV13name);

					renderers       = gameObject.GetComponents<Renderer>();
					rendererToggles = new JSONStorableBool[renderers.Count()];
					counter         = 0;
					foreach (Renderer r in renderers)
					{
						string idString = TYPE_RENDERER_ID;
						if (renderers.Count() > 1) { idString += (counter+1); }
						rendererToggles[counter] = new JSONStorableBool(idString + "_" + prefixedName, r.enabled, (bool _val) => UpdateGameObject());
						m_toggles.Add(rendererToggles[counter]);
						m_preV13toggleNames.Add(idString + "_" + preV13name);
						counter++;
					}

					colliders       = gameObject.GetComponents<Collider>();
					colliderToggles = new JSONStorableBool[colliders.Count()];
					counter         = 0;
					foreach (Collider c in colliders)
					{
						string idString = TYPE_COLLIDER_ID;
						if (colliders.Count() > 1) { prefix += (counter+1); }
						colliderToggles[counter] = new JSONStorableBool(idString + "_" + prefixedName, c.enabled, (bool _val) => UpdateGameObject());
						m_toggles.Add(colliderToggles[counter]);
						m_preV13toggleNames.Add(idString + "_" + preV13name);
						counter++;
					}

					audioSources       = gameObject.GetComponents<AudioSource>();
					audioSourceToggles = new JSONStorableBool[audioSources.Count()];
					counter            = 0;
					foreach (AudioSource a in audioSources)
					{
						string idString = TYPE_AUDIOSOURCE_ID;
						if (audioSources.Count() > 1) { prefix += (counter+1); }
						audioSourceToggles[counter] = new JSONStorableBool(idString + "_" + prefixedName, a.enabled, (bool _val) => UpdateGameObject());
						m_toggles.Add(audioSourceToggles[counter]);
						m_preV13toggleNames.Add(idString + "_" + preV13name);
						counter++;
					}

					light           = gameObject.GetComponent<Light>();
					camera          = gameObject.GetComponent<Camera>();
					lightProbeGroup = gameObject.GetComponent<LightProbeGroup>();
					reflectionProbe = gameObject.GetComponent<ReflectionProbe>();

					// make the toggles public
					foreach (var toggle in m_toggles)
					{
						script.RegisterBool(toggle);
					}
				}
			}


			public void DetermineBounds()
			{	
				bounds = new Bounds();
				bool first = true;
				
				foreach (Collider c in colliders)
				{
					if (c.enabled && c.gameObject.activeInHierarchy) // disabled colliders have zero bounds > will misrepresent the BB
					{
						if (first) { bounds.SetMinMax(c.bounds.min, c.bounds.max); first = false; }
						else       { bounds.Encapsulate(c.bounds); }
					}
				}
				
				foreach (Renderer r in renderers)
				{
					if (first) { bounds.SetMinMax(r.bounds.min, r.bounds.max); first = false; }
					else       { bounds.Encapsulate(r.bounds); }
				}

				if (first)
				{
					// no renderers or colliders > make a 100 unit high thin box
					Vector3 pos = this.gameObject.transform.position;
					bounds.SetMinMax(pos, pos);
					bounds.Encapsulate(pos + 100 * Vector3.up);
					bounds.Expand(0.01f);
				}
			}

			public void DeregisterComponents()
			{
				foreach (var toggle in m_toggles)
				{
					script.DeregisterBool(toggle);
				}
				gameObjectToggle   = null;
				rendererToggles    = null;
				colliderToggles    = null;
				audioSourceToggles = null;
				m_toggles.Clear();
				m_preV13toggleNames.Clear();
			}
								
			
			public string TypeToString()
			{
				string strOut = "";
				string glue   = "";

				if (renderers.Count()    >  0) { strOut += glue + renderers.Count()    + " Renderer"    + (renderers.Count()    == 1 ? "" : "s"); glue = ", "; }
				if (colliders.Count()    >  0) { strOut += glue + colliders.Count()    + " Collider"    + (colliders.Count()    == 1 ? "" : "s"); glue = ", "; }
				if (audioSources.Count() >  0) { strOut += glue + audioSources.Count() + " AudioSource" + (audioSources.Count() == 1 ? "" : "s"); glue = ", "; }
				if (light             != null) { strOut += glue + "Light";                       glue = ", "; }
				if (camera            != null) { strOut += glue + "Camera";                      glue = ", "; }
				if (lightProbeGroup   != null) { strOut += glue + "LightProbeGroup";             glue = ", "; }
				if (reflectionProbe   != null) { strOut += glue + "ReflectionProbe";             glue = ", "; }
				
				if (glue.Length > 0) { strOut = "[" +strOut + "]"; }

				return strOut;
			}


			public int CreateFilteredUI(string typeFilter, ElementNameFilter nameFilter)
			{
				if ( ( (typeFilter == TYPE_GAMEOBJECTS     )                              ) ||  
					 ( (typeFilter == TYPE_LIGHTS          ) && (light           != null) ) ||
					 ( (typeFilter == TYPE_CAMERAS         ) && (camera          != null) ) ||
					 ( (typeFilter == TYPE_LIGHTPROBES     ) && (lightProbeGroup != null) ) ||
					 ( (typeFilter == TYPE_REFLECTIONPROBES) && (reflectionProbe != null) ) )
				{
					// no specific toggle for those (as there should at most be only one per GameObject) > control them via the GameObject
					if (nameFilter.Passes(gameObjectToggle)) { CreateToggle(gameObjectToggle); }
				}
				else if (typeFilter == TYPE_RENDERERS)
				{
					foreach(var rt in rendererToggles)
					{
						if (nameFilter.Passes(rt)) { CreateToggle(rt); }
					}
				}
				else if (typeFilter == TYPE_COLLIDERS)
				{
					foreach(var ct in colliderToggles)
					{
						if (nameFilter.Passes(ct)) { CreateToggle(ct); }
					}
				}
				else if (typeFilter == TYPE_AUDIOSOURCES)
				{
					foreach(var at in audioSourceToggles)
					{
						if (nameFilter.Passes(at)) { CreateToggle(at); }
					}
				}

				AdjustForLongNames();
				AddOnHoverHandlers();
				UpdateUI();
				return m_uiElements.Count();
			}


			protected void CreateToggle(JSONStorableBool _parameter)
			{
				UIDynamicToggle uiTgl = script.CreateToggle(_parameter, true);
				m_uiElements.Add(uiTgl);
			}


			public void UpdateGameObject()
			{
				if (!destroyed && script.enabled)
				{
					gameObject.SetActive(gameObjectToggle.val);
					for (int i=0; i < renderers.Count();    i++) { renderers[i].enabled    = rendererToggles[i].val;    }
					for (int i=0; i < colliders.Count();    i++) { colliders[i].enabled    = colliderToggles[i].val;    }
					for (int i=0; i < audioSources.Count(); i++) { audioSources[i].enabled = audioSourceToggles[i].val; }
				}

				UpdateUI();
			}


			protected void AdjustForLongNames()
			{
				foreach (UIDynamicToggle uiToggle in m_uiElements)
				{
					int len = uiToggle.label.Length;
					uiToggle.labelText.resizeTextForBestFit = true;
					uiToggle.labelText.resizeTextMinSize    = 20;
					uiToggle.labelText.resizeTextMaxSize    = 28;
				}
			}


			protected void AddOnHoverHandlers()
			{
				foreach (UIDynamicToggle uiToggle in m_uiElements)
				{
					Toggle tgl = uiToggle.toggle;
					OnHoverHandler ohh = tgl.gameObject.AddComponent<OnHoverHandler>();
					ohh.SetGameObject(this);
				}
			}


			public void UpdateUI()
			{
				foreach (UIDynamicToggle tgl in m_uiElements) 
				{
					if (script.m_showPrefixes.val && (prefix != "")) { tgl.label = prefix + "_" + name; }
					else                                             { tgl.label = name; }

					// set colour based on toggle/destroyed state
					if (destroyed)            { tgl.backgroundColor = COLOUR_DESTROYED; } // mark destroyed elements specifically
					else if (tgl.toggle.isOn) { tgl.backgroundColor = COLOUR_ENABLED;   } // otherwise by toggle value
					else                      { tgl.backgroundColor = COLOUR_DISABLED;  }

					// special case: element enabled, but game object is not
					if (!destroyed && tgl.toggle.isOn && !gameObjectToggle.val) { tgl.backgroundColor = COLOUR_DISABLED_GO; }
				}
			}


			public void RestoreGameObject()
			{
				if (!destroyed)
				{					
					gameObject.SetActive(gameObjectToggle.defaultVal);
					for (int i=0; i < renderers.Count();    i++) { renderers[i].enabled    = rendererToggles[i].defaultVal;    }
					for (int i=0; i < colliders.Count();    i++) { colliders[i].enabled    = colliderToggles[i].defaultVal;    }
					for (int i=0; i < audioSources.Count(); i++) { audioSources[i].enabled = audioSourceToggles[i].defaultVal; }
				}
			}


			public void DestroyUI()
			{
				foreach (var uiElement in m_uiElements)
				{
					script.RemoveSpacer(uiElement);
				}
				m_uiElements.Clear();
			}


			public void EnableElements()
			{
				foreach (var uiElement in m_uiElements) { uiElement.toggle.isOn = true; }
			}


			public void DisableElements()
			{
				foreach (var uiElement in m_uiElements) { uiElement.toggle.isOn = false; }
			}

			public void RestoreElements()
			{
				foreach (var toggle in m_toggles) { toggle.SetValToDefault(); };
			}


			public void DestroyIfDisabled()
			{
				if (!gameObjectToggle.val) 
				{ 
					Destroy(gameObject);
					destroyed = true;
					UpdateUI();
				}
			}


			public void SaveConfiguration(JSONClass _configurationNode)
			{
				if (_configurationNode == null) return;
				foreach (var toggle in m_toggles) 
				{ 
					_configurationNode[toggle.name] = new JSONData(toggle.val); 
				}
			}


			public void LoadConfiguration(JSONClass _configurationNode)
			{
				if (_configurationNode == null) return;

				for (int idx = 0 ; idx < m_toggles.Count() ; idx++) 
				{ 
					JSONStorableBool toggle = m_toggles[idx];
					if (_configurationNode.HasKey(toggle.name))
					{
						toggle.val = _configurationNode[toggle.name].AsBool;
					}
					else if (_configurationNode.HasKey(m_preV13toggleNames[idx]))
					{
						// for pre-v1.3 backwards compatibility
						toggle.val = _configurationNode[m_preV13toggleNames[idx]].AsBool;
					}
				}
			}


			public void HilightElement(bool _hilight)
			{
				script.ClearHilightedElement();
				if (_hilight) 
				{ 
					script.SetHilightedElement(this); 
				}
			}


			public void SetColor(Color _col)
			{
				bool needToStore = m_colors.Count == 0;
  				foreach (Renderer r in renderers)
				{
					Material[] materials = r.materials;
					for (int idx = 0; idx < materials.Length; idx++)
					{
						if (needToStore) { m_colors.Add(materials[idx].color); }
						materials[idx].color = _col;
					}
					r.materials = materials;
				}
			}


			public void RestoreColors()
			{
				foreach (Renderer r in renderers)
				{
					Material[] materials = r.materials;
					for (int idx = 0; idx < materials.Length; idx++)
					{
						materials[idx].color = m_colors[0];
						m_colors.RemoveAt(0);
					}
					r.materials = materials;
				}
				m_colors.Clear();
			}


			public float DistanceTo(Vector3 worldPosition)
			{
				return Vector3.Distance(worldPosition, bounds.ClosestPoint(worldPosition));
			}
		}


		protected class ElementNameFilter
		{
			public ElementNameFilter(string nameFilter, bool useRegex)
			{
				this.nameFilter = nameFilter;

				regex      = null;
				regexError = false;
				if (useRegex && (nameFilter.Length > 0))
				{
					try
					{
						regex = new Regex(nameFilter);
					}
					catch (Exception e)
					{
						regexError = true;
						regex = null;
					}
				}
			}

			public bool Passes(JSONStorableBool element)
			{
				return regex == null ? element.name.ToLower().Contains(nameFilter.ToLower()) : regex.IsMatch(element.name);
			}


			public bool HasRegexError() 
			{
				return regexError;
			}

			protected string nameFilter;
			protected Regex  regex;
			protected bool   regexError;
		}


		protected class OnHoverHandler : UIBehaviour, IPointerEnterHandler, IPointerExitHandler
		{
			public void SetGameObject(CUA_GameObject cuaGameObject)
			{
				m_gameObject = cuaGameObject;
			}

			public void OnPointerEnter(PointerEventData eventData)
			{
				m_gameObject.HilightElement(true);
			}

			public void OnPointerExit(PointerEventData eventData)
			{
				m_gameObject.HilightElement(false);
			}

			protected CUA_GameObject m_gameObject;
		}


		public override void Init()
		{
			try
			{
				Utils.OnInitUI(CreateUIElement);

				m_cuaGameObjects = new List<CUA_GameObject>();
				m_configuration  = null;

				m_hilightedElement = null;
				m_boundsRenderer   = null;

				Atom atom = GetContainingAtom();
				if (!atom.type.Equals("CustomUnityAsset"))
				{
					logError("Atom is not a Custom Unity Asset (CUA).");
					this.enabled = false;
					return;
				}
				
				m_cuaLoader = FindCustomUnityAssetLoader(atom);
				if (m_cuaLoader == null)
				{
					logError("Unable to find CustomUnityAssetLoader to register callback to.");
					return;
				}
				m_cuaLoader.RegisterAssetLoadedCallback(OnCUALoaded);
				m_cuaLoader.RegisterAssetClearedCallback(OnCUACleared);

				SuperController.singleton.onSceneLoadedHandlers += OnSceneLoaded;

				CreateBaseUI();

				ScanCUA();
			}
			catch (Exception e) { logError($"{e}"); }
		}


		protected void CreateBaseUI()
		{
			CreateHeader($"CUA Editor v{PLUGIN_VERSION}", false, 40, 50);
				
			CreateSpacer().height = 10;
			CreateHeader("Filtering:", false);

			m_cuaElementTypeFilter = new JSONStorableStringChooser("typeFilter", TYPE_OPTIONS, TYPE_OPTIONS[0], "Type", (string _val) => ApplyFilters());
			m_cuaElementTypeFilter.isStorable   = false;
			m_cuaElementTypeFilter.isRestorable = false;
			CreatePopup(m_cuaElementTypeFilter, false);

			m_cuaElementNameFilter = new JSONStorableString("nameFilter", "", (string _val) => ApplyFilters());
			m_cuaElementNameFilter.isStorable   = false;
			m_cuaElementNameFilter.isRestorable = false;
			UIDynamicInputXButton filterByNameInput = Utils.SetupInputXButton(this, m_cuaElementNameFilter, () => { m_cuaElementNameFilter.val = ""; ApplyFilters(); }, false);

			m_useRegexToFilter = new JSONStorableBool("Use RegEx to Filter", false, (bool _val) => ApplyFilters());
			m_useRegexToFilter.isStorable   = false;
			m_useRegexToFilter.isRestorable = false;
			var tglUseRegexToFilter = CreateToggle(m_useRegexToFilter, false);

			CreateSpacer().height = 10;
			CreateHeader("Actions:", false);

			m_showPrefixes = new JSONStorableBool("Show UID Prefixes", false, (bool _val) => UpdateUI());
			m_cuaElementTypeFilter.isStorable   = false;
			m_cuaElementTypeFilter.isRestorable = false;
			var gtlShowPrefixes = CreateToggle(m_showPrefixes, false);

			UIDynamicTwinButton twinButtonEnableDisable = Utils.SetupTwinButton(this,
				"Enable All", EnableAllElements,
				"Disable All", DisableAllElements,
				false);
			ColorBlock cb = twinButtonEnableDisable.buttonLeft.colors;  cb.normalColor = COLOUR_ENABLED;  twinButtonEnableDisable.buttonLeft.colors  = cb;
			           cb = twinButtonEnableDisable.buttonRight.colors; cb.normalColor = COLOUR_DISABLED; twinButtonEnableDisable.buttonRight.colors = cb;

			CreateSpacer().height = 5;
			
			var buttonPrintStructure = CreateButton("Copy CUA Structure to Clipboard", false);
			buttonPrintStructure.button.onClick.AddListener(CopyScenegraphToClipboard);

			var buttonSaveConfiguration = CreateButton("Save Configuration to Clipboard", false);
			buttonSaveConfiguration.button.onClick.AddListener(SaveConfigurationToClipboard);

			var buttonLoadConfiguration = CreateButton("Load Configuration from Clipboard", false);
			buttonLoadConfiguration.button.onClick.AddListener(LoadConfigurationFromClipboard);

			var btnRescan = CreateButton("Rescan CUA", false);
			btnRescan.button.onClick.AddListener(RescanCUA);

			CreateSpacer().height = 10;
			CreateHeader("Destruction:", false);

			m_destroyOnLoad = new JSONStorableBool("Destroy on Load", false);
			var tglDestroyOnLoad = CreateToggle(m_destroyOnLoad, false);
			tglDestroyOnLoad.backgroundColor = COLOUR_DESTROYED;
			RegisterBool(m_destroyOnLoad);

			var buttonExecuteDestroy = CreateButton("Destroy Disabled GameObjects", false);
			buttonExecuteDestroy.buttonColor = COLOUR_DESTROYED;
			buttonExecuteDestroy.button.onClick.AddListener(DestroyDisabledGameObjects);
		}
		

		protected void Update()
		{
			if (m_hilightedElement != null) // that's the only thing we render
			{
				Color c = Color.HSVToRGB(Mathf.PingPong(Time.unscaledTime, 1), 1, 1); 
				m_hilightedElement.SetColor(c);

				if (m_boundsRenderer != null)
				{
					m_boundsRenderer.SetColor(c);
					m_boundsRenderer.SetWidth(m_hilightedElement.DistanceTo(Camera.main.transform.position) * 0.001f);
				}
			}
		}

		
		protected void UpdateUI()
		{
			foreach (CUA_GameObject go in m_cuaGameObjects) { go.UpdateUI(); }
		}


		protected void ApplyFilters()
		{
			StartCoroutine(ApplyFiltersCoroutine());
		}


		protected IEnumerator ApplyFiltersCoroutine()
		{
			yield return new WaitForEndOfFrame();
			DestroyUI();
			yield return new WaitForEndOfFrame();
			BuildUI();
		}


		protected void BuildUI()
		{
			int uiElementCount = 0;
			ElementNameFilter nameFilter = new ElementNameFilter(m_cuaElementNameFilter.val, m_useRegexToFilter.val);
			if (nameFilter.HasRegexError()) { logError($"RegEx 'm_cuaElementNameFilter.val' does not compile."); }
			
			foreach (CUA_GameObject go in m_cuaGameObjects)
			{
				uiElementCount += go.CreateFilteredUI(m_cuaElementTypeFilter.val, nameFilter);
			}
			
			if (uiElementCount == 0)
			{
				m_noElementsFoundLabel = Utils.SetupInfoOneLine(this, $"No {m_cuaElementTypeFilter.val} found", true);
			}
		}
		

		protected void DestroyUI()
		{
			ClearHilightedElement();
			foreach (var go in m_cuaGameObjects)
			{
				go.DestroyUI();
			}
			if (m_noElementsFoundLabel != null) { RemoveSpacer(m_noElementsFoundLabel); }
		}


		protected CustomUnityAssetLoader FindCustomUnityAssetLoader(Atom atom)
		{
			// first, try to find CUA_Loader directly
			CustomUnityAssetLoader cl = atom
				.GetComponentsInChildren<CustomUnityAssetLoader>(true)
					.Where(loader => loader.containingAtom == atom).FirstOrDefault();

			if ((cl == null) && (atom.parentAtom != null))
			{
				// not found > maybe because this atom is parented?
				cl = atom.transform.parent
					.GetComponentsInChildren<CustomUnityAssetLoader>(true)
						.Where(loader => loader.containingAtom == atom).FirstOrDefault();
			
				if ((cl == null) && (atom.parentAtom.isSubSceneType))
				{
					// still not found > maybe because it's a subscene?
					cl = atom.parentAtom
						.GetComponentsInChildren<CustomUnityAssetLoader>(true)
							.Where(loader => loader.containingAtom == atom).FirstOrDefault();
				}
			}

			return cl;
		}


		public void EnableAllElements()
		{
			foreach (CUA_GameObject go in m_cuaGameObjects) { go.EnableElements(); }
		}
		

		public void DisableAllElements()
		{
			foreach (CUA_GameObject go in m_cuaGameObjects) { go.DisableElements(); }
		}


		public void RestoreAllElements()
		{
			foreach (CUA_GameObject go in m_cuaGameObjects) { go.RestoreElements(); }
		}


		public void DestroyDisabledGameObjects()
		{
			foreach (CUA_GameObject go in m_cuaGameObjects) { go.DestroyIfDisabled(); }
		}


		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			// store data for when asset is actually loaded
			m_configuration = jc;
		}
		
	   
		protected void OnCUALoaded()
		{
			ScanCUA(); 
			if (m_destroyOnLoad.val)
			{
				DestroyDisabledGameObjects();
			}   
		}


		protected void OnCUACleared()
		{
			DestroyUI();
			DeregisterAndClearGameObjects();
		}


		protected void OnSceneLoaded()
		{
			// nothing here yet            
		}


		public void RescanCUA()
		{
			m_configuration = null; // don't apply anything from before
			ScanCUA();
		}


		protected void ScanCUA()
		{
			if ((m_cuaLoader != null) && m_cuaLoader.isAssetLoaded)
			{
				DestroyUI();
				DeregisterAndClearGameObjects();

				RecursiveScanCUA(m_cuaLoader.transform, "");

				foreach (CUA_GameObject go in m_cuaGameObjects)
				{
					go.LoadConfiguration(m_configuration);
				}

				BuildUI();
			}
		}


		protected void RecursiveScanCUA(Transform transform, string prefix)
		{
			if (transform != null)
			{
				if (!transform.gameObject.name.Equals("rescaleObject")) // no need to show the VAM CUA root object
				{ 
					CUA_GameObject cuaGameObject = new CUA_GameObject(this, transform, prefix);
					m_cuaGameObjects.Add(cuaGameObject);
				}

				int    childCount   = transform.childCount;
				string numberFormat = "D" + (1 + (int) Mathf.Log10(childCount));
				for (int childIdx = 0; childIdx < transform.childCount; childIdx++)
				{
					// append to prefix only if more than one child (only then could there be duplicates)
					string childPrefix = (childCount > 1) ? (prefix + childIdx.ToString(numberFormat)) : prefix;
					RecursiveScanCUA(transform.GetChild(childIdx), childPrefix);
				}
			}
		}


		protected void CopyScenegraphToClipboard()
		{
			if (m_cuaLoader != null)
			{
				StringBuilder sb = new StringBuilder();
				sb.Append("CUA scenegraph\n");
				RecursivePrintScenegraph(m_cuaLoader.transform, "", sb);
				logDebug(sb.ToString());
				GUIUtility.systemCopyBuffer = sb.ToString();
			}
		}


		protected void RecursivePrintScenegraph(Transform transform, string parentInfo, StringBuilder sb)
		{
			if (transform != null)
			{
				CUA_GameObject cuaGameObject = new CUA_GameObject(this, transform, "");
				
				string indent = parentInfo.Replace("0", "  ") // no more children
					.Replace("1", "│")  // vertical line
					.Replace("2", "├")  // middle child
					.Replace("3", "└"); // last child

				indent += transform.childCount > 0 ? "┬" : "─"; // only child, or beginning of many?

				sb.Append(indent).Append(' ').Append(cuaGameObject.name).Append(' ').Append(cuaGameObject.TypeToString()).Append('\n');
				
				parentInfo = parentInfo.Replace("3", "0").Replace("2", "1"); // last child > keep gaps from here onwards, otherwise a vertical bar

				for (int idx = 0 ; idx < transform.childCount ; idx++)
				{
					string addendum = (idx == transform.childCount-1) ? "3" : "2";
					RecursivePrintScenegraph(transform.GetChild(idx), parentInfo + addendum, sb);
				}
			}
		}


		public void DeregisterAndClearGameObjects()
		{
			foreach (var go in m_cuaGameObjects)
			{
				go.DeregisterComponents();
			}
			m_cuaGameObjects.Clear();
		}


		public void SaveConfigurationToClipboard()
		{
			JSONClass configuration = new JSONClass();
			foreach (CUA_GameObject go in m_cuaGameObjects)
			{
				go.SaveConfiguration(configuration);
			}
			GUIUtility.systemCopyBuffer = configuration.ToString("");
		}


		public void LoadConfigurationFromClipboard()
		{
			string strConfig = GUIUtility.systemCopyBuffer;
			JSONClass configuration = JSON.Parse(strConfig).AsObject;
			foreach (CUA_GameObject go in m_cuaGameObjects)
			{
				go.LoadConfiguration(configuration);
			}
			GUIUtility.systemCopyBuffer = configuration.ToString("");
		}


		protected void SetHilightedElement(CUA_GameObject _element)
		{
			if (m_hilightedElement != null)
			{
				ClearHilightedElement();
			}

			m_hilightedElement = _element;
			
			if (m_boundsRenderer == null) 
			{
				try
				{
				m_boundsRenderer = new DebugLineRenderer("CUA_Editor_Debug_Bounds", 16);
				} catch (Exception e) { logError(e.ToString()); }
			}

			m_hilightedElement.DetermineBounds(); // in case transform has changed
			int     idx = 0;
			Vector3 min = m_hilightedElement.bounds.min;
			Vector3 max = m_hilightedElement.bounds.max;
			m_boundsRenderer.SetPoint(idx++, min.x, min.y, min.z);
			m_boundsRenderer.SetPoint(idx++, min.x, max.y, min.z);
			m_boundsRenderer.SetPoint(idx++, min.x, max.y, max.z);
			m_boundsRenderer.SetPoint(idx++, min.x, min.y, max.z);
			m_boundsRenderer.SetPoint(idx++, max.x, min.y, max.z);
			m_boundsRenderer.SetPoint(idx++, max.x, max.y, max.z);
			m_boundsRenderer.SetPoint(idx++, max.x, max.y, min.z);
			m_boundsRenderer.SetPoint(idx++, max.x, min.y, min.z);
			m_boundsRenderer.SetPoint(idx++, min.x, min.y, min.z);
			m_boundsRenderer.SetPoint(idx++, min.x, min.y, max.z);
			m_boundsRenderer.SetPoint(idx++, min.x, max.y, max.z);
			m_boundsRenderer.SetPoint(idx++, max.x, max.y, max.z);
			m_boundsRenderer.SetPoint(idx++, max.x, min.y, max.z);
			m_boundsRenderer.SetPoint(idx++, max.x, min.y, min.z);
			m_boundsRenderer.SetPoint(idx++, max.x, max.y, min.z);
			m_boundsRenderer.SetPoint(idx++, min.x, max.y, min.z);
			m_boundsRenderer.FinalisePoints();
		}


		protected void ClearHilightedElement()
		{
			if (m_boundsRenderer != null)
			{
				m_boundsRenderer.Destroy();
				m_boundsRenderer = null;
			}
			if (m_hilightedElement != null)
			{
				m_hilightedElement.RestoreColors();
				m_hilightedElement = null;
			}
		}


		
		// ********************************************************************
		#region Logging
		
		protected void logDebug(string text) 
		{
			SuperController.LogMessage($"{nameof(CUA_Editor)}: {text}");
		}
		
		protected void logError(string text) 
		{
			SuperController.LogError($"{nameof(CUA_Editor)}: {text}");
		}

		#endregion Logging

	
		// ********************************************************************
		#region Utilities
		
		protected void CreateHeader(string header, bool rightSide, int fontSize = 32, int fieldHeight = 40) 
		{
			JSONStorableString staticDescString      = new JSONStorableString($"HDR_{header}", $"<size={fontSize}><b>{header}</b></size>");
			UIDynamicTextField staticDescStringField = CreateTextField(staticDescString, rightSide);
			LayoutElement sdsfLayout  	 = staticDescStringField.GetComponent<LayoutElement>();
			sdsfLayout.preferredHeight   = sdsfLayout.minHeight = fieldHeight;
			staticDescStringField.height          = fieldHeight;
			staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
		}


		#endregion Utilities


		// ********************************************************************
		#region Enable, Disable, Destroy
		
		protected void OnEnable()
		{
			foreach (CUA_GameObject go in m_cuaGameObjects) { go.UpdateGameObject(); }
		}


		protected void OnDisable() 
		{
			foreach (CUA_GameObject go in m_cuaGameObjects) { go.RestoreGameObject(); }
		}


		protected void OnDestroy()
		{
			if (SuperController.singleton != null) 
			{
				SuperController.singleton.onSceneLoadedHandlers -= OnSceneLoaded;
			}
			Utils.OnDestroyUI();
			ClearHilightedElement();
		}
		

		#endregion Enable, Disable, Destroy
	
		protected JSONStorableBool          m_showPrefixes;
		protected JSONStorableBool          m_destroyOnLoad;
		protected CustomUnityAssetLoader    m_cuaLoader;
		protected List<CUA_GameObject>      m_cuaGameObjects;
		protected JSONClass                 m_configuration;
		
		protected JSONStorableStringChooser m_cuaElementTypeFilter;
		protected JSONStorableString        m_cuaElementNameFilter;
		protected JSONStorableBool          m_useRegexToFilter;
		protected UIDynamicTextInfo         m_noElementsFoundLabel;

		protected CUA_GameObject            m_hilightedElement;
		protected DebugLineRenderer         m_boundsRenderer;
	}
}
